package stream;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import se.rupy.http.*;

public class Server extends Service implements Node, Runnable {
	ConcurrentHashMap list;	// The pull streams.
	Thread thread;			// To purge "noop" messages so proxies don't timeout the connections.
	Node node;				// The routing implementation.

	/* Deprecated: before I used JSON Array as transport.
	 * JSON Array is uncommented in the client to remove 
	 * lib dependency so don't change this.
	 */
	boolean atomic = true;	// Atomic or JSON Array transport?
	boolean debug = false;
	boolean alive;

	private Queue find(String name) {
		Iterator it = list.values().iterator();

		while(it.hasNext()) {
			Queue queue = (Queue) it.next();

			if(queue.name.equals(name)) {
				return queue;
			}
		}

		return null;
	}

	public String path() {
		return "/push:/pull";
	}

	public void create(Daemon daemon) throws Exception {
		if(!alive) {
			list = new ConcurrentHashMap();
			alive = true;

			thread = new Thread(this);
			thread.start();

			node = new Test();
			node.call(this);
		}
	}

	public void run() {
		try {
			while(alive) {
				purge(false);
				Thread.sleep(5000);
			}

			if(debug)
				System.err.println("purge");

			purge(true);
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}

	public void destroy() throws Exception {
		alive = false;
	}

	public void call(Node node) {
		// we'll call you, don't call us!
	}

	public void push(String name, String message) {
		Queue queue = find(name);

		if(queue != null) {
			queue.add(message);

			int wakeup = queue.event.reply().wakeup();

			if(wakeup == Reply.CLOSED || wakeup == Reply.COMPLETE) {
				if(debug)
					System.err.println(queue.event.index() + " remove 1");
				list.remove(queue.event);
			}
		}
	}

	private void purge(boolean finish) throws IOException {
		Iterator it = list.values().iterator();

		while(it.hasNext()) {
			Queue queue = (Queue) it.next();

			queue.add("noop");

			if(finish) {
				if(debug)
					System.err.println(queue.event.index() + " finish 1");
				queue.finish = true;
			}

			int wakeup = queue.event.reply().wakeup();

			if(wakeup == Reply.CLOSED || wakeup == Reply.COMPLETE) {
				if(debug)
					System.err.println(queue.event.index() + " remove 2");
				it.remove();
			}
		}
	}

	public void filter(Event event) throws Event, Exception {
		if(event.query().path().equals("/push")) {
			event.query().parse();

			String name = event.string("name");
			String message = event.string("message");

			long time = System.currentTimeMillis();
			
			node.push(name, message);
			
			time = System.currentTimeMillis() - time;

			if(time > 10) {
				System.out.println("too long " + time);
			}

			event.output().print("");
			throw event;
		}

		if(event.push()) {
			try {
				Queue queue = (Queue) list.get(event);
				Output out = event.output();

				if(queue == null || queue.finish) {
					if(debug)
						System.err.println(event.index() + " finish 2");
					out.finish();
				}

				if(queue != null && !queue.isEmpty()) {
					if(atomic) { // Atomic transport
						String message = (String) queue.poll();

						while(message != null) {
							String accept = queue.event.query().header("accept");
							
							if(accept != null && accept.equals("text/event-stream")) {
								out.print("data: " + message + "\n\n");
							}
							else {
								out.write((message + "\n").getBytes());
							}
							//System.out.print(".");
							message = (String) queue.poll();
						}
						//System.out.println("");
					}
					else { // JSON Array transport
						/* Deprecated: before I used JSON Array as transport.
						 */
						boolean first = true;
						out.print("[");

						String message = (String) queue.poll();

						while(message != null) {
							if(first) {
								first = false;
							}
							else {
								out.print(",");
							}

							out.print("\"" + message + "\"");

							message = (String) queue.poll();
						}

						out.print("]\0");
					}
					
					out.flush();
				}
			}
			catch(Exception e) {
				if(debug)
					System.err.println(event.index() + " remove 3");
				list.remove(event);
				throw e;
			}
		}
		else {
			event.query().parse();
			String name = event.string("name");
			Queue queue = find(name);

			if(queue != null) {
				if(debug)
					System.err.println(event.index() + " remove 4");
				list.remove(queue.event);
			}

			list.put(event, new Queue(name, event));

			String accept = event.query().header("accept");
			
			if(accept != null && accept.equals("text/event-stream")) {
				event.reply().type("text/event-stream");
			}
			
			event.hold();
			Output out = event.output();
			
			if(atomic) { // Atomic transport
				if(accept != null && accept.equals("text/event-stream")) {
					out.print("data: noop\n\n");
				}
				else {
					out.print("noop\n");
				}
			}
			else { // JSON Array transport
				/* Deprecated: before I used JSON Array as transport.
				 */
				out.print("[\"noop\"]\0");
			}
			
			out.flush();
		}
	}

	/*
	 * The queue contains the event and the messages 
	 * to be sent to the client.
	 */
	static class Queue extends ConcurrentLinkedQueue {
		private String name;
		private Event event;
		private boolean finish;

		public Queue(String name, Event event) {
			this.name = name;
			this.event = event;
		}
	}
	
	/*
	 * This whole thing is bogus, the port you will use needs to respond 
	 * to this anyhow, so 843 is a complete waste of everyones time.
	 * 
	public static class Policy extends Service implements Runnable {
		ServerSocket server;
		boolean alive;
		
		public String path() {
			return "/policy";
		}

		public void create(Daemon daemon) throws Exception {
			if(server == null) {
				alive = true;
				server = new ServerSocket(843);
				new Thread(this).start();
			}
		}
		
		public void destroy() throws Exception {
			alive = false;
			server.close();
		}
		
		public void run() {
			try {
				while(alive) {
					Socket socket = server.accept();
					BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
					PrintWriter out = new PrintWriter(socket.getOutputStream(), true);
					
					String line = in.readLine();
					
					if(line.equals("<policy-file-request/>\0")) {
						System.out.println("policy-yo");
						out.println("<cross-domain-policy><allow-access-from domain=\"*\" to-ports=\"*\"/></cross-domain-policy>\0");
					}
					
					Thread.currentThread().sleep(10);
					
					socket.close();
				}
			}
			catch(SocketException e) {
				// Never Mind
			}
			catch(Exception e) {
				e.printStackTrace();
			}
		}
		
		public void filter(Event event) throws Event, Exception {}
	}
	*/
}